@@ -2,7 +2,7 @@
 
 /*
     libzint - the open source barcode library
-    Copyright (C) 2009 - 2020 Robin Stuart <rstuart114@gmail.com>
+    Copyright (C) 2009 - 2021 Robin Stuart <rstuart114@gmail.com>
 
     Redistribution and use in source and binary forms, with or without
     modification, are permitted provided that the following conditions
@@ -31,8 +31,8 @@
  */
 /* vim: set ts=4 sw=4 et : */
 
+#include <errno.h>
 #include <stdio.h>
-#include <string.h>
 #include "common.h"
 #include <math.h>
 #ifdef _MSC_VER
@@ -43,13 +43,8 @@
 
 #define SSET    "0123456789ABCDEF"
 
-/* Index of transparent color, -1 for no transparent color
- * This might be set into a variable if transparency is activated as an option
- */
-#define TRANSPARENT_INDEX (-1)
-
 typedef struct s_statestruct {
-    unsigned char * pOut;
+    unsigned char *pOut;
     unsigned char *pIn;
     unsigned int InLen;
     unsigned int OutLength;
@@ -86,7 +81,7 @@ static unsigned char NextPaletteIndex(statestruct *pState)
 }
 
 
-static char BufferNextByte(statestruct *pState) {
+static int BufferNextByte(statestruct *pState) {
     (pState->OutPosCur)++;
     /* Check if this position is a byte count position
      * fg_f_bytecountbyte_set indicates, if byte count position bytes should be
@@ -107,11 +102,10 @@ static char BufferNextByte(statestruct *pState) {
     return 0;
 }
 
-static char AddCodeToBuffer(statestruct *pState, unsigned short CodeIn, unsigned char CodeBits) {
+static int AddCodeToBuffer(statestruct *pState, unsigned short CodeIn, unsigned char CodeBits) {
     /* Check, if we may fill up the current byte completely */
     if (CodeBits >= pState->OutBitsFree) {
-        (pState->pOut)[pState->OutPosCur] |= (unsigned char)
-                (CodeIn << (8 - pState->OutBitsFree));
+        (pState->pOut)[pState->OutPosCur] |= (unsigned char) (CodeIn << (8 - pState->OutBitsFree));
         if (BufferNextByte(pState))
             return -1;
         CodeIn = (unsigned short) (CodeIn >> pState->OutBitsFree);
@@ -128,8 +122,7 @@ static char AddCodeToBuffer(statestruct *pState, unsigned short CodeIn, unsigned
     }
     /* The remaining bits of CodeIn fit in the current byte. */
     if (CodeBits > 0) {
-        (pState->pOut)[pState->OutPosCur] |= (unsigned char)
-            (CodeIn << (8 - pState->OutBitsFree));
+        (pState->pOut)[pState->OutPosCur] |= (unsigned char) (CodeIn << (8 - pState->OutBitsFree));
         pState->OutBitsFree -= CodeBits;
     }
     return 0;
@@ -154,7 +147,7 @@ static unsigned short FindPixelOutlet(statestruct *pState, unsigned short HeadNo
     return 0;
 }
 
-static char NextCode(statestruct *pState, unsigned char * pPixelValueCur, unsigned char CodeBits) {
+static int NextCode(statestruct *pState, unsigned char *pPixelValueCur, unsigned char CodeBits) {
     unsigned short UpNode;
     unsigned short DownNode;
     /* start with the root node for last pixel chain */
@@ -208,9 +201,9 @@ static int gif_lzw(statestruct *pState, int paletteBitSize) {
         paletteBitSize = 2;
 
     /* initial size of compression codes */
-    CodeBits = paletteBitSize+1;
+    CodeBits = paletteBitSize + 1;
     pState->ClearCode = (1 << paletteBitSize);
-    pState->FreeCode = pState->ClearCode+2;
+    pState->FreeCode = pState->ClearCode + 2;
     pState->OutBitsFree = 8;
     pState->OutPosCur = -1;
     pState->fByteCountByteSet = 0;
@@ -237,7 +230,7 @@ static int gif_lzw(statestruct *pState, int paletteBitSize) {
         return 0;
 
     for (;;) {
-        char Res;
+        int Res;
         /* generate and save the next code, which may consist of multiple input pixels. */
         Res = NextCode(pState, &PixelValueCur, CodeBits);
         if (Res < 0)
@@ -254,7 +247,8 @@ static int gif_lzw(statestruct *pState, int paletteBitSize) {
             }
             // > Update last bytecount byte;
             if (pState->OutByteCountPos < pState->OutPosCur) {
-                (pState->pOut)[pState->OutByteCountPos] = (unsigned char) (pState->OutPosCur - pState->OutByteCountPos - 1);
+                (pState->pOut)[pState->OutByteCountPos]
+                    = (unsigned char) (pState->OutPosCur - pState->OutByteCountPos - 1);
             }
             pState->OutPosCur++;
             return pState->OutPosCur;
@@ -278,7 +272,7 @@ static int gif_lzw(statestruct *pState, int paletteBitSize) {
 /*
  * Called function to save in gif format
  */
-INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
+INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, unsigned char *pixelbuf) {
     unsigned char outbuf[10];
     FILE *gif_file;
     unsigned short usTemp;
@@ -286,24 +280,29 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
     int colourCount;
     unsigned char paletteRGB[10][3];
     int paletteCount, paletteCountCur, paletteIndex;
-    int pixelIndex;
+    unsigned int pixelIndex;
     int paletteBitSize;
     int paletteSize;
     statestruct State;
+    int transparent_index;
+    int bgindex = -1, fgindex = -1;
+    const int output_to_stdout = symbol->output_options & BARCODE_STDOUT;
 
     unsigned char backgroundColourIndex;
     unsigned char RGBCur[3];
+    unsigned char RGBUnused[3] = {0,0,0};
 
     int colourIndex;
 
     int fFound;
 
     unsigned char pixelColour;
+    unsigned int bitmapSize = symbol->bitmap_height * symbol->bitmap_width;
 
     /* Allow for overhead of 4 == code size + byte count + overflow byte + zero terminator */
-    unsigned int lzoutbufSize = symbol->bitmap_height * symbol->bitmap_width + 4;
+    unsigned int lzoutbufSize = bitmapSize + 4;
 #ifdef _MSC_VER
-    char * lzwoutbuf;
+    char *lzwoutbuf;
 #endif
 
 #ifndef _MSC_VER
@@ -312,22 +311,6 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
     lzwoutbuf = (char *) _alloca(lzoutbufSize);
 #endif /* _MSC_VER */
 
-    /* Open output file in binary mode */
-    if ((symbol->output_options & BARCODE_STDOUT) != 0) {
-#ifdef _MSC_VER
-        if (-1 == _setmode(_fileno(stdout), _O_BINARY)) {
-            strcpy(symbol->errtxt, "610: Can't open output file");
-            return ZINT_ERROR_FILE_ACCESS;
-        }
-#endif
-        gif_file = stdout;
-    } else {
-        if (!(gif_file = fopen(symbol->outfile, "wb"))) {
-            strcpy(symbol->errtxt, "611: Can't open output file");
-            return ZINT_ERROR_FILE_ACCESS;
-        }
-    }
-
     /*
      * Build a table of the used palette items.
      * Currently, there are the following 10 colour codes:
@@ -362,8 +345,7 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
     colourCount = 0;
     paletteCount = 0;
     /* loop over all pixels */
-    for ( pixelIndex = 0; pixelIndex < (symbol->bitmap_height * symbol->bitmap_width); pixelIndex++)
-    {
+    for (pixelIndex = 0; pixelIndex < bitmapSize; pixelIndex++) {
         fFound = 0;
         /* get pixel colour code */
         pixelColour = pixelbuf[pixelIndex];
@@ -416,7 +398,7 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
                 RGBCur[0] = 0; RGBCur[1] = 0; RGBCur[2] = 0;
                 break;
             default: /* error case - return  */
-                strcpy(symbol->errtxt, "611: unknown pixel colour");
+                strcpy(symbol->errtxt, "612: unknown pixel colour");
                 return ZINT_ERROR_INVALID_DATA;
         }
         /* Search, if RGB value is already present */
@@ -439,6 +421,9 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
             paletteRGB[paletteIndex][2] = RGBCur[2];
 
             paletteCount++;
+
+            if (pixelColour == '0') bgindex = paletteIndex;
+            if (pixelColour == '1') fgindex = paletteIndex;
         }
         /* Add palette index to current colour code */
         (State.colourCode)[colourCount] = pixelColour;
@@ -447,11 +432,28 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
     }
     State.colourCount = colourCount;
 
+    /* Set transparency */
+    /* Note: does not allow both transparent foreground and background -
+     * background takes priority */
+    transparent_index = -1;
+    if (strlen(symbol->fgcolour) > 6) {
+        if ((symbol->fgcolour[6] == '0') && (symbol->fgcolour[7] == '0')) {
+            // Transparent foreground
+            transparent_index = fgindex;
+        }
+    }
+    if (strlen(symbol->bgcolour) > 6) {
+        if ((symbol->bgcolour[6] == '0') && (symbol->bgcolour[7] == '0')) {
+            // Transparent background
+            transparent_index = bgindex;
+        }
+    }
+
     /* find palette bit size from palette size*/
 
     /* 1,2 -> 1, 3,4 ->2, 5,6,7,8->3 */
     paletteBitSize = 0;
-    paletteCountCur = paletteCount-1;
+    paletteCountCur = paletteCount - 1;
     while (paletteCountCur != 0) {
         paletteBitSize++;
         paletteCountCur >>= 1;
@@ -461,11 +463,27 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
         paletteBitSize = 1;
 
     /* palette size 2 ^ bit size */
-    paletteSize = 1<<paletteBitSize;
+    paletteSize = 1 << paletteBitSize;
+
+    /* Open output file in binary mode */
+    if (output_to_stdout) {
+#ifdef _MSC_VER
+        if (-1 == _setmode(_fileno(stdout), _O_BINARY)) {
+            sprintf(symbol->errtxt, "610: Could not set stdout to binary (%d: %.30s)", errno, strerror(errno));
+            return ZINT_ERROR_FILE_ACCESS;
+        }
+#endif
+        gif_file = stdout;
+    } else {
+        if (!(gif_file = fopen(symbol->outfile, "wb"))) {
+            sprintf(symbol->errtxt, "611: Could not open output file (%d: %.30s)", errno, strerror(errno));
+            return ZINT_ERROR_FILE_ACCESS;
+        }
+    }
 
     /* GIF signature (6) */
     memcpy(outbuf, "GIF87a", 6);
-    if (TRANSPARENT_INDEX != -1)
+    if (transparent_index != -1)
         outbuf[4] = '9';
 
     fwrite(outbuf, 6, 1, gif_file);
@@ -508,15 +526,14 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
     fwrite(paletteRGB, 3*paletteCount, 1, gif_file);
     /* add unused palette items to fill palette size */
     for (paletteIndex = paletteCount; paletteIndex < paletteSize; paletteIndex++) {
-        unsigned char RGBCur[3] = {0,0,0};
-        fwrite(RGBCur, 3, 1, gif_file);
+        fwrite(RGBUnused, 3, 1, gif_file);
     }
 
     /* Graphic control extension (8) */
     /* A graphic control extension block is used for overlay gifs.
      * This is necessary to define a transparent color.
      */
-    if (TRANSPARENT_INDEX != -1) {
+    if (transparent_index != -1) {
         /* Extension Introducer = '!' */
         outbuf[0] = '\x21';
         /* Graphic Control Label */
@@ -534,7 +551,7 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
         outbuf[4] = 0;
         outbuf[5] = 0;
         /* Transparent Color Index */
-        outbuf[6] = (unsigned char) TRANSPARENT_INDEX;
+        outbuf[6] = (unsigned char) transparent_index;
         /* Block Terminator */
         outbuf[7] = 0;
         fwrite(outbuf, 8, 1, gif_file);
@@ -563,22 +580,29 @@ INTERNAL int gif_pixel_plot(struct zint_symbol *symbol, char *pixelbuf) {
     fwrite(outbuf, 10, 1, gif_file);
 
     /* prepare state array */
-    State.pIn = (unsigned char *) pixelbuf;
-    State.InLen = symbol->bitmap_height * symbol->bitmap_width;
+    State.pIn = pixelbuf;
+    State.InLen = bitmapSize;
     State.pOut = (unsigned char *) lzwoutbuf;
     State.OutLength = lzoutbufSize;
 
     /* call lzw encoding */
     byte_out = gif_lzw(&State, paletteBitSize);
     if (byte_out <= 0) {
-        fclose(gif_file);
+        if (!output_to_stdout) {
+            fclose(gif_file);
+        }
+        strcpy(symbol->errtxt, "613: Insufficient memory for LZW buffer");
         return ZINT_ERROR_MEMORY;
     }
     fwrite(lzwoutbuf, byte_out, 1, gif_file);
 
     /* GIF terminator */
     fputc('\x3b', gif_file);
-    fclose(gif_file);
+    if (output_to_stdout) {
+        fflush(gif_file);
+    } else {
+        fclose(gif_file);
+    }
 
     return 0;
 }
